@@ -10,13 +10,21 @@
 #include <NTL/ctools.h>
 #include <NTL/new.h>
 
+#include <utility>
 #include <iostream>
 #include <new>
 #include <stdexcept>
+#include <streambuf>
 
 #include <cstdlib>
 #include <cmath>
+#include <cstring>
 
+#ifdef NTL_SAFE_VECTORS
+
+#include <type_traits>
+
+#endif
 
 
 #if (defined(NTL_THREADS) && defined(__GNUC__) && !defined(NTL_DISABLE_TLS_HACK))
@@ -35,6 +43,23 @@
 #define NTL_SNS std ::
 #define NTL_USE_SNS using namespace std;
 
+#define NTL_IMPORT_FROM_STD \
+   using NTL_SNS abs; \
+   using NTL_SNS ceil; \
+   using NTL_SNS exp; \
+   using NTL_SNS fabs; \
+   using NTL_SNS floor; \
+   using NTL_SNS ldexp; \
+   using NTL_SNS log; \
+   using NTL_SNS sqrt; \
+   using NTL_SNS ostream;  \
+   using NTL_SNS istream;  \
+   using NTL_SNS cerr;  \
+   using NTL_SNS ifstream;  \
+   using NTL_SNS ofstream; 
+
+
+
 #ifndef NTL_LEGACY_NO_NAMESPACE
 
 // This wraps NTL in the NTL namespace.
@@ -50,15 +75,8 @@
 // functions that are both overloaded by NTL and are used in
 // the implementation of NTL.
 
-#define NTL_START_IMPL NTL_USE_SNS NTL_OPEN_NNS \
-   using NTL_SNS abs; \
-   using NTL_SNS ceil; \
-   using NTL_SNS exp; \
-   using NTL_SNS fabs; \
-   using NTL_SNS floor; \
-   using NTL_SNS ldexp; \
-   using NTL_SNS log; \
-   using NTL_SNS sqrt; 
+#define NTL_START_IMPL NTL_OPEN_NNS NTL_IMPORT_FROM_STD
+
 
 #define NTL_END_IMPL NTL_CLOSE_NNS
 
@@ -73,7 +91,7 @@
 #define NTL_USE_NNS 
 #define NTL_NNS 
 
-#define NTL_START_IMPL
+#define NTL_START_IMPL NTL_IMPORT_FROM_STD
 #define NTL_END_IMPL
 
 #endif
@@ -332,24 +350,6 @@ inline void conv(unsigned long& x, float a) { x = ((unsigned int) to_long(a)); }
 inline void conv(unsigned long& x, double a) { x = ((unsigned int) to_long(a)); }
 
 
-/* ------------------------------------- */
-
-
-// new style converson function
-//   example: ZZ x = conv<ZZ>(1);
-//   note: modern C++ compilers should implemented 
-//     "named return value optimization", so the
-//     result statement should not create a temporary
-
-template<class T, class S>
-T conv(const S& a)
-{
-   T x;
-   conv(x, a);
-   return x;
-}
-
-
 // some convenience casting routines:
 
 inline long cast_signed(unsigned long a) { return long(a); }
@@ -457,8 +457,6 @@ void PrintTime(NTL_SNS ostream& s, double t);
 // semantics.
 
 // A call to Deleter::apply should free the pointed-to storage
-// and set the pointer itself to zero, so apply should
-// take an argument that is a reference to a T*.
 
 template<class T, class Deleter>
 class WrappedPtr {
@@ -473,7 +471,7 @@ public:
    WrappedPtr() : rep(0) { }
    void operator=(const raw_ptr& _rep)  { rep = _rep; }
 
-   ~WrappedPtr() { Deleter::apply(rep); } 
+   ~WrappedPtr() { if (rep) Deleter::apply(rep); } 
 
    operator const raw_ptr& () const { return rep; }
    operator raw_ptr& () { return rep; }
@@ -481,10 +479,17 @@ public:
    const raw_ptr* operator&() const { return &rep; }
    raw_ptr* operator&() { return &rep; }
 
-   void kill() { Deleter::apply(rep); }
+   void kill() { if (rep) { Deleter::apply(rep); rep = 0; } }
 
    void swap(WrappedPtr& other) { _ntl_swap(rep, other.rep); }
 
+   void move(WrappedPtr& other) 
+   {
+      WrappedPtr tmp;
+      tmp.swap(other);
+      tmp.swap(*this);
+   }
+
 };
 
 template<class T, class Deleter>
@@ -616,8 +621,12 @@ operator+(scope_guard_builder b, F&& f)
 
 
 class DummyScopeGuard {
+  bool active;
 public:
-   void relax() { }
+   DummyScopeGuard() : active(true) { }
+   ~DummyScopeGuard() { if (active) TerminalError("unexpected exception"); }
+   
+   void relax() { active = false; }
 };
 
 #define NTL_SCOPE(var) DummyScopeGuard var; if (false)
@@ -686,7 +695,15 @@ struct key_wrapper {
 
 #else
 
-#define NTL_TLS_LOCAL_INIT(type,var,init) static NTL_THREAD_LOCAL type var init
+
+// NOTE: this definition of NTL_TLS_LOCAL_INIT ensures that var names
+// a local reference, regardless of the implementation
+#define NTL_TLS_LOCAL_INIT(type,var,init) \
+    static NTL_THREAD_LOCAL type _ntl_hidden_variable_tls_local ## var init; \
+    type &var = _ntl_hidden_variable_tls_local ## var
+
+
+
 
 #endif
 
@@ -706,7 +723,434 @@ struct key_wrapper {
 
 #define NTL_TLS_GLOBAL_ACCESS(var) \
 _ntl_hidden_typedef_tls_access_ ## var & var = _ntl_hidden_function_tls_access_ ## var()
+
+
+// **************************************************************
+// Following is code for "long long" arithmetic that can
+// be implemented using NTL_ULL_TYPE or using assembly.
+// I have found that the assembly can be a bit faster.
+// For now, this code is only available if NTL_HAVE_LL_TYPE
+// is defined.  This could change.  In any case, this provides
+// a cleaner interface and might eventually allow for 
+// implementation on systems that don't provide a long long type.
+// **************************************************************
+
+#ifdef NTL_HAVE_LL_TYPE
+
       
+#if (!defined(NTL_DISABLE_LL_ASM) \
+     && defined(__GNUC__) && (__GNUC__ >= 4) && !defined(__INTEL_COMPILER)  && !defined(__clang__) \
+     && defined (__x86_64__)  && NTL_BITS_PER_LONG == 64)
+
+// NOTE: clang's and icc's inline asm code gen is pretty bad, so
+// we don't even try. 
+
+// FIXME: probably, this should all be properly tested for speed (and correctness)
+// using the Wizard.
+
+
+struct ll_type {
+   unsigned long hi, lo;
+};
+
+
+inline void 
+ll_mul_add(ll_type& x, unsigned long a, unsigned long b)
+{
+  unsigned long hi, lo;
+   __asm__ (
+   "mulq %[b] \n\t" 
+   "addq %[lo],%[xlo] \n\t"
+   "adcq %[hi],%[xhi]" : 
+   [lo] "=a" (lo), [hi] "=d" (hi), [xhi] "+r" (x.hi), [xlo] "+r" (x.lo) : 
+   [a] "%[lo]" (a), [b] "rm" (b) :
+   "cc"
+   );
+}
+
+inline void 
+ll_imul_add(ll_type& x, unsigned long a, unsigned long b)
+{
+  unsigned long hi, lo;
+   __asm__ (
+   "imulq %[b] \n\t" 
+   "addq %[lo],%[xlo] \n\t"
+   "adcq %[hi],%[xhi]" : 
+   [lo] "=a" (lo), [hi] "=d" (hi), [xhi] "+r" (x.hi), [xlo] "+r" (x.lo) : 
+   [a] "%[lo]" (a), [b] "rm" (b) :
+   "cc"
+   );
+}
+
+inline void 
+ll_mul(ll_type& x, unsigned long a, unsigned long b)
+{
+   __asm__ (
+   "mulq %[b]" :
+   [lo] "=a" (x.lo), [hi] "=d" (x.hi) : 
+   [a] "%[lo]" (a), [b] "rm" (b) :
+   "cc"
+   );
+}
+
+inline void 
+ll_imul(ll_type& x, unsigned long a, unsigned long b)
+{
+   __asm__ (
+   "imulq %[b]" :
+   [lo] "=a" (x.lo), [hi] "=d" (x.hi) : 
+   [a] "%[lo]" (a), [b] "rm" (b) :
+   "cc"
+   );
+}
+
+inline void
+ll_add(ll_type& x, unsigned long a)
+{
+   __asm__ (
+   "addq %[a],%[xlo] \n\t"
+   "adcq %[z],%[xhi]" :
+   [xhi] "+r" (x.hi), [xlo] "+r" (x.lo) :
+   [a] "rm" (a), [z] "i" (0) :
+   "cc"
+   );
+}
+
+
+
+// NOTE: an optimizing compiler will remove the conditional.
+// The alternative would be to make a specialization for shamt=0.
+// Unfortunately, this is impossible to do across a wide range
+// of compilers and still maintain internal linkage --- it is not 
+// allowed to include static spec in the specialization (new compilers
+// will complain) and without it, some older compilers will generate
+// an external symbol.  In fact, NTL currently never calls 
+// this with shamt=0, so it is all rather academic...but I want to
+// keep this general for future use.
+
+// NOTE: this implementation assumes that shamt is in the range 
+// 0..NTL_BITS_PER_LONG-1
+
+#if 1
+
+// The shrd instruction can be very slow on some
+// machines.  Two shifts is usually just as good.
+
+template<long shamt>
+unsigned long
+ll_rshift_get_lo(ll_type x)
+{
+   unsigned long res;
+   if (shamt) 
+      res = (x.lo >> shamt) | (x.hi << (NTL_BITS_PER_LONG-shamt));
+   else
+      res = x.lo;
+      
+   return res;
+}
+
+#else
+
+
+template<long shamt>
+unsigned long
+ll_rshift_get_lo(ll_type x)
+{
+   if (shamt) {
+      __asm__ (
+      "shrdq %[shamt],%[hi],%[lo]" :
+      [lo] "+r" (x.lo) : 
+      [shamt] "i" (shamt), [hi] "r" (x.hi) :
+      "cc"
+      );
+   }
+   return x.lo;
+}
+
+#endif
+
+
+inline unsigned long 
+ll_get_lo(const ll_type& x)
+{
+   return x.lo;
+}
+
+inline unsigned long 
+ll_get_hi(const ll_type& x)
+{
+   return x.hi;
+}
+
+
+inline void
+ll_init(ll_type& x, unsigned long a)
+{
+   x.lo = a;
+   x.hi = 0;
+}
+
+#else
+
+
+typedef NTL_ULL_TYPE ll_type;
+
+// NOTE: the following functions definitions should serve as
+// documentation, as well.
+
+inline void 
+ll_mul_add(ll_type& x, unsigned long a, unsigned long b)
+{
+   x += ((ll_type) a)*((ll_type) b);
+}
+
+// a and b should be representable as positive long's,
+// to allow for the most flexible implementation
+inline void 
+ll_imul_add(ll_type& x, unsigned long a, unsigned long b)
+{
+   x += ((ll_type) long(a))*((ll_type) long(b));
+}
+inline void 
+ll_mul(ll_type& x, unsigned long a, unsigned long b)
+{
+   x = ((ll_type) a)*((ll_type) b);
+}
+
+// a and b should be representable as positive long's,
+// to allow for the most flexible implementation
+inline void 
+ll_imul(ll_type& x, unsigned long a, unsigned long b)
+{
+   x = ((ll_type) long(a))*((ll_type) long(b));
+}
+
+inline void
+ll_add(ll_type& x, unsigned long a)
+{
+   x += a;
+}
+
+// NOTE: shamt must be in the range 0..NTL_BITS_PER_LONG-1
+template<long shamt>
+unsigned long
+ll_rshift_get_lo(const ll_type& x)
+{
+   return ((unsigned long) (x >> shamt));
+}
+
+inline unsigned long 
+ll_get_lo(const ll_type& x)
+{
+   return ((unsigned long) x);
+}
+
+inline unsigned long 
+ll_get_hi(const ll_type& x)
+{
+   return ((unsigned long) (x >> NTL_BITS_PER_LONG));
+}
+
+
+inline void
+ll_init(ll_type& x, unsigned long a)
+{
+   x = a;
+}
+
+
+#endif
+
+
+inline unsigned long 
+ll_mul_hi(unsigned long a, unsigned long b)
+{
+   ll_type x;
+   ll_mul(x, a, b);
+   return ll_get_hi(x);
+} 
+
+
+#endif
+
+
+
+
+
+#ifdef NTL_SAFE_VECTORS
+
+
+#define NTL_RELOC_TAG (relocatable)
+
+#define NTL_DECLARE_RELOCATABLE_WHEN(x) \
+constexpr bool DeclareRelocatableType x
+
+#if (defined(NTL_HAVE_COPY_TRAITS1))
+
+
+// This strategy is used on compilers that fully support C++11 type traits.
+// For some reason, is_trivially_copyable says "true" even if the class
+// has deleted it's copy constructor.  Which means it is not copyable at all.
+// So I added an explicit test for is_copy_constructible.
+// Just to be on the safe side, I check for a trivial destructor.
+
+// This strategy is checked in the CheckCOPY_TRAITS1.cpp program.
+
+
+template<class T>
+constexpr bool Relocate_aux_has_trivial_copy(T*)
+{
+   return  std::is_trivially_copyable<T>::value &&
+           std::is_trivially_destructible<T>::value &&
+           std::is_copy_constructible<T>::value;
+}
+
+#elif (defined(NTL_HAVE_COPY_TRAITS2))
+
+// This strategy is needed on GCC before v5.0, as the required type
+// traits are not impplemented.  Note that on a class with it's copy
+// constructors deleted, __has_trivial_copy is false on GCC before 4.9
+// and true startig with 4.9.  So I had to make use of SFINAE techniques
+// to make sure there actually is a non-deleted copy constructor.
+// Just to be on the safe side, I check for a trivial destructor.
+
+// This strategy is checked in the CheckCOPY_TRAITS1.cpp program.
+
+template <bool statement, typename out>
+struct Relocate_aux_Failable
+{
+     typedef out Type;
+};
+
+struct Relocate_aux_TwoChars { char d[2]; };
+
+template <typename T>
+struct Relocate_aux_has_copy
+{
+
+     static const T *MakeT();
+
+     template <typename U> // U and T are the same type
+     static typename Relocate_aux_Failable<(bool(sizeof U(*MakeT()))), char>::Type copy(int);
+
+     template <typename U>
+     static typename Relocate_aux_Failable<true, Relocate_aux_TwoChars>::Type copy(...);
+
+     enum { value =  sizeof( copy<T>(0) )  == 1 };
+};
+
+
+template<class T>
+constexpr bool Relocate_aux_has_trivial_copy(T*)
+{
+   return  __has_trivial_copy(T) &&
+           __has_trivial_destructor(T) &&
+           Relocate_aux_has_copy<T>::value;
+}
+
+
+#else
+
+#error "lacking compiler support for NTL_SAFE_VECTORS"
+
+#endif
+
+
+// NOTE: I've checked the correctness of the above strategies using
+// Godbolt's compiler explorer across a range of complilers
+// (clang, gcc, icc, MS).
+
+
+template<class T>
+constexpr bool DeclareRelocatableType(T*)
+{
+   return Relocate_aux_has_trivial_copy((T*)0);
+}
+ 
+
+#else
+
+#define NTL_RELOC_TAG (true)
+
+#define NTL_DECLARE_RELOCATABLE_WHEN(x) \
+inline bool DeclareRelocatableType x
+
+
+template<class T>
+inline bool DeclareRelocatableType(T*)
+{
+   return true;
+}
+
+#endif
+
+
+#define NTL_DECLARE_RELOCATABLE(x) NTL_DECLARE_RELOCATABLE_WHEN(x) \
+   { return true; }
+
+
+// Examples: 
+//   NTL_DECLARE_RELOCATABLE((int*))
+//   NTL_DECLARE_RELOCATABLE((Foo<int>*)) 
+//   template <class X, class Y> NTL_DECLARE_RELOCATABLE((Foo<X,Y>*))
+
+
+#if (NTL_CXX_STANDARD >= 2011)
+
+#define NTL_DEFAULT =default;
+
+#else
+
+#define NTL_DEFAULT {}
+
+#endif
+
+
+// The following idea for deriving from streambuf comes from:
+// https://stackoverflow.com/questions/1448467/initializing-a-c-stdistringstream-from-an-in-memory-buffer/1449527#1449527
+
+struct plain_c_string_streambuf : public std::streambuf
+{
+   plain_c_string_streambuf(const char* ss)
+   {
+      char *s = const_cast<char*>(ss);
+      // casting away constant should be safe here,
+      // based of my reading of the functionality
+      // of streambuf from the documentation at cplusplus.com.
+    
+      setg(s, s, s + std::strlen(s));
+   }
+};
+
+// Generic conversion from char* or const char*.  We use SFINAE
+// to prevent conversions from 0. 
+
+
+template<class S, class T>
+typename _ntl_enable_if<_ntl_is_char_pointer<T>::value,void>::type
+conv(S& x, T y)
+{
+   if (!y) InputError("bad conversion from char*");
+   plain_c_string_streambuf buf(y);
+   std::istream istr(&buf);
+   if (!(istr >> x)) InputError("bad conversion from char*");
+}
+
+
+
+// new style converson function
+//   example: ZZ x = conv<ZZ>(1);
+//   note: modern C++ compilers should implemented 
+//     "named return value optimization", so the
+//     result statement should not create a temporary
+
+template<class T, class S>
+T conv(const S& a)
+{
+   T x;
+   conv(x, a);
+   return x;
+}
 
 
 
